#!/usr/bin/python
# -*- coding: utf-8 -*-

import jinja2
from . import templates

__all__ = ['get_template']

__text_templates = {
    'cbf_header': templates.cbf_header,
    'edf_header': templates.edf_header,
    'esp_header': templates.esp_header,
    'parfile': templates.pafile,
    'aliases': templates.dectrisaliases_ini,
    'monitor': templates.dectris_monitor,
    'set2M': templates.setfile2M,
    'set6M': templates.setfile6M,
    'set300K': templates.setfile300K,
    'set1M': templates.setfile1M,
    'setFrelon': templates.setfileFrelon,
    'CrysalisExpSettings.ini': templates.crysalisExpSettings_ini,
}

__bin_templates = {
    'sel_od': templates.sel_od,
    'sel_odEsp': templates.sel_odEsp,
    'ccd1m': templates.ccd1M,
    'ccd2m': templates.ccd2M,
    'ccd6m': templates.ccd6M,
    'ccdesperanto': templates.ccdEsp,
}

__env = jinja2.Environment(loader=jinja2.DictLoader(__text_templates), newline_sequence='\r\n')


def get_template(name):
    if name in __text_templates:
        return __env.get_template(name)
    elif name in __bin_templates:
        return __bin_templates[name]
    raise KeyError(f'There is no template with name {name}')



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Fork of fabio edf reader
"""

import mmap
import numpy as np
from . import cbfimage, templates, cryioimage


class EdfError(cryioimage.ImageError):
    pass


class EdfImage(cryioimage.CryioImage):
    Extensions = '.edf',
    EDF_HEADER_BEGIN = b'{'
    EDF_HEADER_END = b'}'
    DTYPE = {  # borrowed from fabio
        "SignedByte": np.int8,
        "Signed8": np.int8,
        "UnsignedByte": np.uint8,
        "Unsigned8": np.uint8,
        "SignedShort": np.int16,
        "Signed16": np.int16,
        "UnsignedShort": np.uint16,
        "Unsigned16": np.uint16,
        "UnsignedShortInteger": np.uint16,
        "SignedInteger": np.int32,
        "Signed32": np.int32,
        "UnsignedInteger": np.uint32,
        "Unsigned32": np.uint32,
        "SignedLong": np.int32,
        "UnsignedLong": np.uint32,
        "Signed64": np.int64,
        "Unsigned64": np.uint64,
        "FloatValue": np.float32,
        "FLOATVALUE": np.float32,
        "FLOAT": np.float32,  # fit2d
        "Float": np.float32,  # fit2d
        "FloatIEEE32": np.float32,
        "Float32": np.float32,
        "Double": np.float64,
        "DoubleValue": np.float64,
        "FloatIEEE64": np.float64,
        "DoubleIEEE64": np.float64
    }
    TYPE2STR = {
        np.dtype('int8'): "SignedByte",
        np.dtype('int16'): "SignedShort",
        np.dtype('int32'): "SignedInteger",
        np.dtype('int64'): "Signed64",
        np.dtype('uint8'): "UnsignedByte",
        np.dtype('uint16'): "UnsignedShort",
        np.dtype('uint32'): "UnsignedInteger",
        np.dtype('uint64'): "Unsigned64",
        np.dtype('float32'): "FloatValue",
        np.dtype('float64'): "DoubleValue",
        # np.dtype('float128'): "QuadrupleValue",  # it does not exist in win32 numpy
    }

    def __init__(self, filepath=None):
        super().__init__(filepath)
        self._offset = 0
        if self.filepath:
            self._read()

    def _read(self):
        with open(self.filepath, 'rb') as f:
            try:
                self._map = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            except (mmap.error, IOError):
                self._map = f.read()
        self._seek_header()
        if self._begin == -1 or self._end == -1:
            raise EdfError('Could not find header tokens in {}. Might it be not an EDF?'.format(self.filepath))
        self._parse()
        self.multiframe = self._begin != -1 and self._end != -1

    def __iter__(self):
        while True:
            yield self
            if self._begin == -1 or self._end == -1:
                raise StopIteration
            self._parse()

    def _seek_header(self):
        self._begin = self._map.find(self.EDF_HEADER_BEGIN, self._offset) + 1
        self._end = self._map.find(self.EDF_HEADER_END, self._offset)

    def _parse(self):
        self._parse_header()
        self._parse_binary()
        self._seek_header()

    def _get_chunk(self, start, size):
        end = start + size
        self._offset = end
        return self._map[start:end]

    def _parse_header(self):
        raw = self._get_chunk(self._begin, self._end-self._begin).decode()
        self.header = {}
        for line in raw.splitlines():
            try:
                key, value = line.strip(' ;').split('=')
            except ValueError:
                continue
            value = value.strip().split(' ')
            key = key.strip()
            if not value:
                continue
            elif len(value) == 1:
                value = value[0]
                try:
                    self.header[key] = float(value) if '.' in value else int(value)
                except ValueError:
                    self.header[key] = value
            else:
                try:
                    self.header[key] = [float(i) if '.' in i else int(i) for i in value]
                except ValueError:
                    self.header[key] = value
        if not self.header:
            raise EdfError('Could not parse EDF header or it is empty in {}'.format(self.filepath))
        # Some files do not contains "Size" key, we have to get the type here to calculate size correctly
        self._parse_type()
        if 'Size' not in self.header:
            if 'Dim_1' in self.header and 'Dim_2' in self.header:
                self.header['Size'] = self.header['Dim_1'] * self.header['Dim_2'] * self.header['ntype'].itemsize
            else:
                raise EdfError('The EDF image header of {0} does not contain "Size" key.'.format(self.filepath))

    def _parse_type(self):
        try:
            t = self.DTYPE[self.header['DataType']]
        except KeyError:
            t = np.int32
        dt = np.dtype(t)
        if 'ByteOrder' in self.header and self.header['ByteOrder'] == 'HighByteFirst':
            dt.newbyteorder('>')
        self.header['ntype'] = dt

    def _parse_binary(self,):
        binsize = int(self.header['Size'])
        start = self._end + 2  # }\n
        raw = self._get_chunk(start, binsize)
        self.array = np.frombuffer(raw, self.header['ntype']).reshape((self.header['Dim_2'], self.header['Dim_1']))
        self.array.flags.writeable = True

    def save(self, filepath, fmt='cbf', **kwargs):
        try:
            {
                'cbf': self.save_cbf,
                'edf': self.save_edf,
            }[fmt](filepath, **kwargs)
        except KeyError:
            raise EdfError('Format is unknown')

    def save_cbf(self, filepath, **kwargs):
        cbf = cbfimage.CbfImage()
        cbf.array = self.array
        cbf.header.update(kwargs)
        cbf.save_cbf(filepath, **kwargs)

    def save_edf(self, filepath, **kwargs):
        if self.array is None:
            return
        shapeX, shapeY = self.array.shape
        binary_blob = self.array.reshape((shapeX, shapeY)).tostring()
        edf = self.header
        edf.update({
            'binsize': len(binary_blob),
            'headersize': 2560,
            'dim_1': shapeY,
            'dim_2': shapeX,
            'datatype': self.TYPE2STR[self.array.dtype]
        })
        edf.update(kwargs)
        header = templates.get_template('edf_header').render(edf)
        header = '{}\n'.format(header[:-2]).encode('ascii', errors='ignore')
        with open(filepath, 'wb') as fedf:
            fedf.write(header)
            fedf.write(binary_blob)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import numpy as np


def padArray(array, x, y, right=True, bottom=True):
    a = np.empty((x - array.shape[0], array.shape[1]), dtype=array.dtype)
    a.fill(-1)
    array = np.vstack((array, a) if right else (a, array))
    a = np.empty((array.shape[0], y - array.shape[1]), dtype=array.dtype)
    a.fill(-1)
    array = np.hstack((array, a) if bottom else (a, array))
    return array



#!/usr/bin/python
# -*- coding: utf-8 -*-


import math
from datetime import datetime
import numpy as np
from . import _cryio, ImageError, tools, templates, cryioimage


class EsperantoException(ImageError):
    pass


class EsperantoImage(cryioimage.CryioImage):
    ESPERANTO_SIZE = 254  # -len('\r\n')

    def __init__(self, filepath=None):
        super().__init__(filepath)
        if self.filepath:
            self.parse()

    def parse(self):
        self.parse_header()
        self.parse_binary()

    def parse_header(self):
        raise NotImplementedError

    def parse_binary(self):
        raise NotImplementedError

    def save(self, filepath, array=None, pack=True, **kwargs):
        if array is None and self.array is None:
            raise EsperantoException('Nothing to save, an array is not provided')
        if array is None:
            array = self.array
        # crysalis reads only squared esperanto arrays with the x,x size divided by 4
        if array.shape[0] != array.shape[1] or array.shape[0] % 4 and array.shape[1] % 4:
            x = y = int(math.ceil(max(array.shape) / 4.0) * 4)
            array = tools.padArray(array, x, y)
        shape = array.shape[0]
        if pack:
            datatype = 'AGI_BITFIELD'
            espstr = bytes(_cryio._esp_encode(np.ascontiguousarray(array, np.int32)))
        else:
            datatype = '4BYTE_LONG'
            espstr = array.reshape((shape ** 2,)).astype(np.int32).tobytes()
        esp_header = {
            'shape': shape,
            'datetime': datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'),
            'datatype': datatype,
        }
        esp_header.update(kwargs)
        header = templates.get_template('esp_header').render(esp_header).splitlines()
        binblob = [bytes(l.ljust(self.ESPERANTO_SIZE).encode('ascii')) for l in header]
        with open(filepath, 'wb') as fedf:
            fedf.write(b'\r\n'.join(binblob))
            fedf.write(b'\r\n')
            fedf.write(espstr)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import numpy as np
from .cryioimage import ImageError, CryioImage


class NotNumpyMask(ImageError):
    pass


class NumpyMask(CryioImage):
    Extensions = '.npz',

    def __init__(self, filepath=None):
        super().__init__(filepath)
        if self.filepath:
            self._read()

    def _read(self):
        try:
            self.array = np.load(self.filepath)['arr_0']
        except TypeError:
            raise NotNumpyMask(f'File "{self.filepath}" does not contain a numpy mask array')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import mmap
import ctypes
import struct
import numpy as np
from . import _cryio, cryioimage

MAR_MODE_DOSE = 0
MAR_MODE_TIME = 1
MAR_FORMAT_COMPRESSED = 1
MAR_FORMAT_SPIRAL = 2
MAR_MAGIC_NUMBER = 1234
MAR_BIN_HEADER_SIZE = 64
MAR_HEADER_SIZE = 4096
MAR_MAGIC_STRING1 = 'CCP4 packed image, X: {:04d}, Y: {:04d}\n'


class Mar345Error(cryioimage.ImageError):
    pass


class _LittleEndianBinaryHeader(ctypes.Structure):
    _format_ = '<16i'
    _fields_ = [
        ('magic_number', ctypes.c_int),
        ('dim1', ctypes.c_int),
        ('n_overflow', ctypes.c_int),
        ('format', ctypes.c_int),
        ('mode', ctypes.c_int),
        ('n_pixels', ctypes.c_int),
        ('pixel_length', ctypes.c_int),
        ('pixel_height', ctypes.c_int),
        ('wavelength', ctypes.c_int),
        ('distance', ctypes.c_int),
        ('start_phi', ctypes.c_int),
        ('end_phi', ctypes.c_int),
        ('start_omega', ctypes.c_int),
        ('end_omega', ctypes.c_int),
        ('chi', ctypes.c_int),
        ('two_theta', ctypes.c_int),
    ]

    def __init__(self, header):
        super(_LittleEndianBinaryHeader, self).__init__(*struct.unpack(self._format_, header))


class _BigEndianBinaryHeader(_LittleEndianBinaryHeader):
    _format_ = '>16i'


class Mar345Image(cryioimage.CryioImage):
    Extensions = '.mar3450', '.mar2300'

    def __init__(self, filepath=None):
        super().__init__(filepath)
        if self.filepath:
            self._read_mmap()
            self._parse_header()
            self._parse_binary()
            self._close_mmap()

    def _parse_header(self):
        self._parse_binary_header()
        self._parse_ascii_header()

    def _parse_binary(self):
        dim1 = self.binary_header.dim1
        dim2 = self.binary_header.n_pixels // dim1
        of = self.binary_header.n_overflow
        if self.binary_header.format == MAR_FORMAT_COMPRESSED:
            fs = MAR_MAGIC_STRING1.format(dim1, dim2).encode()
            s = self._mmap.find(fs)
            if s == -1:
                raise Mar345Error('Only mar v1 can be unpacked, v2 is not yet supported')
            ds = s + len(fs)  # data starts at
            self.array = np.asarray(_cryio._mar_decode(dim1, dim2, of, self._mmap[MAR_HEADER_SIZE:s], self._mmap[ds:]))
        elif self.binary_header.format == MAR_FORMAT_SPIRAL:
            raise Mar345Error('Cannot decode spiral mar format')
        else:
            raise Mar345Error('Mar format is unknown')

    def _parse_binary_header(self):
        self.binary_header = _LittleEndianBinaryHeader(self._mmap[:MAR_BIN_HEADER_SIZE])
        if self.binary_header.magic_number != MAR_MAGIC_NUMBER:
            self.binary_header = _BigEndianBinaryHeader(self._mmap[:MAR_BIN_HEADER_SIZE])

    def _parse_ascii_header(self):
        # ASCII header is absolutely useless, we do not waste time to parse it
        self.header = {}

    def _read_mmap(self):
        with open(self.filepath, 'rb') as f:
            try:
                self._mmap = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            except (mmap.error, IOError):
                self._mmap = f.read()

    def _close_mmap(self):
        if isinstance(self._mmap, mmap.mmap):
            self._mmap.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import cbfimage
from . import edfimage
from . import fit2dmask
from . import mar345image
from . import numpymask
from . import brukerimage
from .cryioimage import ImageError, CryioError


__openable = (
    cbfimage.CbfImage,
    edfimage.EdfImage,
    fit2dmask.Fit2DMask,
    mar345image.Mar345Image,
    numpymask.NumpyMask,
    brukerimage.BrukerImage
)


def imageext():
    for cls in __openable:
        for ext in cls.Extensions:
            yield cls, ext


def extensions():
    for cls, ext in imageext():
        yield ext


def openImage(filename):
    for cls, ext in imageext():
        if filename.endswith(ext):
            return cls(filename)
    raise ImageError(f'CryIO could not recognize the image {filename}')


open_image = openImage



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Fork of fabio fit2d mask reader
"""

import numpy as np
from .cryioimage import ImageError, CryioImage


class NotFit2dMask(ImageError):
    pass


class Fit2DMask(CryioImage):
    HEADER_ID = ('M', 0), ('A', 4), ('S', 8), ('K', 12)
    HEADER_SIZE = 1024
    Extensions = '.msk',

    def __init__(self, filepath=None):
        super().__init__(filepath)
        if self.filepath:
            self.parse()

    def parse(self):
        _map = open(self.filepath, 'rb')
        self.parse_header(_map)
        self.parse_body(_map)
        _map.close()

    def parse_header(self, _map):
        header = _map.read(self.HEADER_SIZE)
        for i, j in self.HEADER_ID:
            c = chr(header[j])
            if c != i:
                raise NotFit2dMask(f'Not a fit2d mask file {self.filepath}')
        fit2dhdr = np.fromstring(header, np.int32)
        self.dim1 = fit2dhdr[4]
        self.dim2 = fit2dhdr[5]

    def parse_body(self, _map):
        num_ints = (self.dim1 + 31) // 32
        total = self.dim2 * num_ints * 4
        data = _map.read(total)
        data = np.fromstring(data, np.uint8).reshape((self.dim2, num_ints * 4))
        # noinspection PyNoneFunctionAssignment
        result = np.empty((self.dim2, num_ints * 4 * 8), np.uint8)
        # Unpack using bitwise comparisons to 2**n
        bits = np.ones(1, np.uint8)
        for i in range(8):
            result[:, i::8] = np.bitwise_and(bits, data).astype(np.uint8)
            bits *= 2
        # Extra rows needed for packing odd dimensions
        spares = num_ints * 4 * 8 - self.dim1
        if spares:
            # noinspection PyTypeChecker
            array = np.where(result[:, :-spares] == 0, 0, 1)
        else:
            # noinspection PyTypeChecker
            array = np.where(result == 0, 0, 1)
        self.array = array.astype(np.int8).reshape((self.dim2, self.dim1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Fork of fabio bruker reader
"""

import os
import mmap
import numpy as np
from .cryioimage import CryioImage, ImageError


class BrukerError(ImageError):
    pass


class BrukerImage(CryioImage):
    Extensions = '.gfrm',
    line = 80
    blocksize = 512
    nhdrblks = 5  # by default we always read 5 blocks of 512
    version = 86
    size = {
        1: np.uint8,
        2: np.uint16,
        4: np.uint32,
    }

    def __init__(self, filepath=None):
        super().__init__(filepath)
        if self.filepath:
            self._read()

    def _read(self):
        with open(self.filepath, 'rb') as f:
            try:
                self._map = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            except (mmap.error, IOError):
                self._map = f
            try:
                self._readheader()
                npixelb = int(self.header['NPIXELB'])
            except (KeyError, ValueError, IndexError):
                raise BrukerImage(f'Bruker file {self.filepath} cannot be read')
            data = np.fromstring(self._map.read(self.dim1 * self.dim2 * npixelb), dtype=self.size[npixelb])
            if not np.little_endian and data.dtype.itemsize > 1:
                data.byteswap(True)

            data = data.astype(np.uint32)
            # handle overflows
            nov = int(self.header['NOVERFL'])
            while nov > 0:  # Read in the overflows
                nov -= 1
                # need at least int32 sized data I guess - can reach 2^21
                # 16 character overflows:
                #      9 characters of intensity
                #      7 character position
                intensity = self._map.read(9)
                position = self._map.read(7)
                data[int(position)] = int(intensity)
            # Handle Float images ...
            if 'LINEAR' in self.header:
                try:
                    slope, offset = map(float, self.header['LINEAR'].split(None, 1))
                    if slope != 1 or offset != 0:
                        data = (data * slope + offset).astype(np.float32)
                except (ValueError, KeyError, IndexError):
                    pass
            self.array = data.reshape(self.dim1, self.dim2)
            self._map = None

    def _readheader(self):
        """
        The bruker format uses 80 char lines in key : value format
        In the first 512*5 bytes of the header there should be a
        HDRBLKS key, whose value denotes how many 512 byte blocks
        are in the total header. The header is always n*5*512 bytes,
        otherwise it wont contain whole key: value pairs
        """
        headerbytes = self._map.read(self.blocksize * self.nhdrblks)
        for i in range(0, self.nhdrblks * self.blocksize, self.line):
            self._fillheader(headerbytes, i)
        # we must have read this in the first 5*512 bytes.
        nhdrblks = int(self.header.get('HDRBLKS', self.nhdrblks))
        self.header['HDRBLKS'] = nhdrblks
        # Now read in the rest of the header blocks, appending
        headerbytes += self._map.read(self.blocksize * (nhdrblks - self.nhdrblks))
        for i in range(self.nhdrblks * self.blocksize, nhdrblks * self.blocksize, self.line):
            self._fillheader(headerbytes, i)
        # make a (new) header item called "datastart"
        self.header['datastart'] = self.blocksize * nhdrblks
        # set the image dimensions
        self.dim1 = int(self.header['NROWS'].split()[0])
        self.dim2 = int(self.header['NCOLS'].split()[0])
        self.version = int(self.header.get('VERSION', self.version))

    def _fillheader(self, headerbytes, i):
        sls = headerbytes[i:i + self.line]
        if b':' in sls:
            key, val = [n.strip().decode('ascii') for n in sls.split(b':', 1)]
            if key in self.header:
                self.header[key] = f'{self.header[key]}{os.linesep}{val}'
            else:
                self.header[key] = val
